/*
   +----------------------------------------------------------------------+
   | PHP Version 5                                                        |
   +----------------------------------------------------------------------+
   | Copyright (c) 1997-2010 The PHP Group                                |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.php.net/license/3_01.txt                                  |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
   | Author: Melanie Rhianna Lewis <cyberspice@php.net>                   |
   +----------------------------------------------------------------------+
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <png.h>

#include "php_framegrab_image.h"

#define FRAMEGRAB_IMAGE_STRING_BUFFER_SIZE 1024

/**
 * Initialises a php_framegrab_image_data structure.
 */
void php_framegrab_image_data_init(php_framegrab_image_data *data) {
	data->bytes_per_line = 0;
	data->format         = FRAMEGRAB_IMAGE_FORMAT_NONE;
	data->width          = 0;
	data->height         = 0;
	data->size           = 0;
	data->image          = NULL;
}

/**
 * Destroys a php_framegrab_image_data structure freeing any memory
 * it is using.
 */
void php_framegrab_image_data_destroy(php_framegrab_image_data *data) {
	if (data->image) {
		free(data->image);
	}
}

/**
 * Clones a php_framegrab_image_data structure.
 */
int php_framegrab_image_data_clone(php_framegrab_image_data *data,
		                           php_framegrab_image_data *copydata) {
	copydata->bytes_per_line = data->bytes_per_line;
	copydata->format         = data->format;
	copydata->width          = data->width;
	copydata->height         = data->height;
	copydata->size           = data->size;

	/* Copy the image pixel data */
	if (data->image) {
		copydata->image = malloc(data->size);
		if (!copydata->image) {
			return 0;
		} else {
			memcpy(copydata->image, data->image, data->size);
		}
	} else {
		copydata->image = NULL;
	}

	return 1;
}

#ifdef FRAMEGRAB_USES_INTEGER_MATH
/**
 * Converts YUV (YCbCr) to RGB (8 bits per colour channel) using the ITU-R standard.  This version
 * uses integer math.  This will be faster on systems that do not have an FPU.
 */
static inline YUVtoRGB888(FG_UINT8 y, FG_UINT8 u, FG_UINT8 v, FG_UINT8 *rgb_ptr) {
	int u1 = ((int)u) - 128;
	int v1 = ((int)v) - 128;
	*rgb_ptr++ = (FG_UINT8)(((int)y) + v1 + v1 >> 2 + v1 >> 3 + v1 >> 5);
	*rgb_ptr++ = (FG_UINT8)(((int)y) - (u1 >> 2 + u1 >> 4 + u1 >> 5) - (v1 >> 1 + v1 >> 3 + v1 >> 4 + v1 >> 5));
	*rgb_ptr   = (FG_UINT8)(((int)y) + u1 + u1 >> 1 + u1 >> 2 + u1 >> 6);
}
#else
/**
 * Converts YUV (YCbCr) to RGB (8 bits per colour channel) using the ITU-R standard.  This version
 * uses floating point math.
 */
static inline YUVtoRGB888(FG_UINT8 y, FG_UINT8 u, FG_UINT8 v, FG_UINT8 *rgb_ptr) {
	float u1 = ((float)u) - 128;
	float v1 = ((float)v) - 128;
	int r1 = (int)((float)y + 1.402 * v1);
	int g1 = (int)((float)y - 0.344 * u1 - 0.714 * v1);
	int b1 = (int)((float)y + 1.722 * u1);
	*rgb_ptr++ = (FG_UINT8)(r1 > 255 ? 255 : r1);
	*rgb_ptr++ = (FG_UINT8)(g1 > 255 ? 255 : g1);
	*rgb_ptr   = (FG_UINT8)(b1 > 255 ? 255 : b1);
}
#endif

/**
 * Converts a YUYV (YUV 4:2:2) pair of pixels to a pair of RGB pixels.
 */
static inline YUYVtoRGB888(FG_UINT8 *yuyv_ptr, FG_UINT8 *rgb_ptr) {
	FG_UINT8 y1 = *yuyv_ptr++;
	FG_UINT8 u  = *yuyv_ptr++;
	FG_UINT8 y2 = *yuyv_ptr++;
	FG_UINT8 v  = *yuyv_ptr;
	YUVtoRGB888(y1, u, v, rgb_ptr);
	YUVtoRGB888(y2, u, v, rgb_ptr + 3);
}

/**
 * Converts a YUYV (YUV 4:2:2) pair of pixels to a pair of RGB pixels.
 */
static inline UYVYtoRGB888(FG_UINT8 *yuyv_ptr, FG_UINT8 *rgb_ptr) {
	FG_UINT8 y1 = *yuyv_ptr++;
	FG_UINT8 u  = *yuyv_ptr++;
	FG_UINT8 y2 = *yuyv_ptr++;
	FG_UINT8 v  = *yuyv_ptr;
	YUVtoRGB888(y1, u, v, rgb_ptr);
	YUVtoRGB888(y2, u, v, rgb_ptr + 3);
}

/**
 * Creates a copy php_framegrab_image_data structure in RGB format
 */
static int framegrab_image_data_reformat_to_RGB888(php_framegrab_image_data *data,
                                                   php_framegrab_image_data *newdata) {
	FG_UINT8 *linptr, *loutptr, *inptr, *outptr;
	int h, w;

	newdata->format = FRAMEGRAB_IMAGE_FORMAT_RGB;

	switch (data->format) {
		case FRAMEGRAB_IMAGE_FORMAT_YUYV:
		case FRAMEGRAB_IMAGE_FORMAT_UYVY:
			newdata->bytes_per_line = newdata->width * 3;
			newdata->size           = newdata->height * newdata->bytes_per_line;
			newdata->image          = malloc(newdata->size);
			if (!newdata) {
				newdata->size = 0;
				return 0;
			}

			linptr  = data->image;
			loutptr = newdata->image;

			/* These cases are very very similar but it would be inefficient
			 * to do a comparison for each line etc.  So we suffer a bit of
			 * code over head to do the comparison outside of the loop. */
			if (FRAMEGRAB_IMAGE_FORMAT_YUYV == data->format) {
				for (h = 0; h < newdata->height; h++) {
					inptr  = linptr;
					outptr = loutptr;
					/* Converting pairs of pixels */
					for (w = 0; w < newdata->width; w += 2) {
						YUYVtoRGB888(inptr, outptr);
						inptr  += 4;
						outptr += 6;
					}
					linptr  += data->bytes_per_line;
					loutptr += newdata->bytes_per_line;
				}
			} else {
				for (h = 0; h < newdata->height; h++) {
					inptr  = linptr;
					outptr = loutptr;
					/* Converting pairs of pixels */
					for (w = 0; w < newdata->width; w += 2) {
						UYVYtoRGB888(inptr, outptr);
						inptr  += 4;
						outptr += 6;
					}
					linptr  += data->bytes_per_line;
					loutptr += newdata->bytes_per_line;
				}
			}
			return 1;

		/* TODO: Other formats */

		default:
			return 0;
	}
}

/**
 * Creates a copy php_framegrab_image_data structure in a new format.
 */
static int framegrab_image_data_reformat(php_framegrab_image_data *data,
		                                 int format,
                                         php_framegrab_image_data *newdata)
{
	if (format == data->format) {
		return php_framegrab_image_data_clone(data, newdata);
	}

	newdata->width  = data->width;
	newdata->height = data->height;

	switch (format) {
		case FRAMEGRAB_IMAGE_FORMAT_RGB:
			return framegrab_image_data_reformat_to_RGB888(data, newdata);

		/* TODO: Other formats */
		default:
			return 0;
	}
}

#define FRAMEGRAB_PNG_WRITE_DATA_CHUNK_SIZE 0x4000

/**
 *
 */
typedef struct _php_framegrab_png_write_buffer_data {
	png_bytep  buf;
	png_size_t buflen;
	png_size_t size;
} php_framegrab_png_write_buffer_data;

/**
 * libpng 'write' function that outputs to a buffer
 */
void framegrab_png_write_data(png_structp png_ptr, png_bytep ptr, png_size_t size) {
	php_framegrab_png_write_buffer_data *data = (php_framegrab_png_write_buffer_data *)png_get_io_ptr(png_ptr);

	/* Hopefully we got the buffer information */
	if (data) {
		/* Do we need to realloc? If so allocate enough chunks to cover the data being
		 * written.*/
		if (data->buflen < (data->size + size)) {
			data->buflen += FRAMEGRAB_PNG_WRITE_DATA_CHUNK_SIZE *
					        ((size / FRAMEGRAB_PNG_WRITE_DATA_CHUNK_SIZE) + 1);
			data->buf = (png_bytep)realloc(data->buf, data->buflen);
		}

		/* Copy the data */
		if (data->buf) {
			/* Calculate the offset in to the buffer each time because the buffer
			 * may have moved in memory. */
			memcpy(data->buf + data->size, ptr, size);
			data->size += size;
		}
	}
}

/**
 * libpng 'flush' function that outputs to a buffer
 */
void framegrab_png_flush_data(png_structp png_ptr) {
	/* Does nothing */
}

/**
 * Writes the image to the specified filename or to memory.  If output to
 * filename set pbuffer and psize to NULL else set filename to NULL.
 */
static int framegrab_image_write_to_png(php_framegrab_image_data *image, char *filename, FG_UINT8 **pbuffer, int *psize) {
	php_framegrab_png_write_buffer_data buffer_data = { NULL, 0, 0};
	FILE *file = NULL;
	png_structp png_ptr = NULL;
	png_infop info_ptr = NULL;
	png_byte **row_ptrs = NULL;
	png_byte *row_ptr = NULL;
	int bit_depth, color_type, h;

	/* Allocate an array of pointers to the rows of the image.  This is used
	 * to write the image to the file. */
	row_ptrs = (png_byte**)calloc(image->height, sizeof(png_byte*));
	if (!row_ptrs) {
		goto png_write_error;
	}

	row_ptr = image->image;
	for (h = 0; h < image->height; h++) {
		row_ptrs[h] = row_ptr;
		row_ptr += image->bytes_per_line;
	}

	if (filename) {
		/* Open the output file */
		file = fopen(filename, "wb");
		if (!file) {
			goto png_write_error;
		}
	} else {
		/* Initialise the write buffer */
		buffer_data.buf = (png_bytep)malloc(FRAMEGRAB_PNG_WRITE_DATA_CHUNK_SIZE);
		if (!buffer_data.buf) {
			goto png_write_error;
		}

		buffer_data.buflen = FRAMEGRAB_PNG_WRITE_DATA_CHUNK_SIZE;
		buffer_data.size   = 0;
	}

	/* Initialise the structures required to write a png */
	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (!png_ptr) {
		goto png_write_error;
	}

	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr) {
		goto png_write_error;
	}

	/* Set up libpng error handling */
	if (setjmp(png_jmpbuf(png_ptr))) {
		goto png_write_error;
	}

	if (file) {
		/* Output to file */
		png_init_io(png_ptr, file);
	} else {
		/* Output to buffer */
		png_set_write_fn(png_ptr, &buffer_data, framegrab_png_write_data, framegrab_png_flush_data);
	}

	/* Set the output format */
	switch (image->format) {
		case FRAMEGRAB_IMAGE_FORMAT_Y8:
			bit_depth = 8;
			color_type = PNG_COLOR_TYPE_GRAY;
			break;

		case FRAMEGRAB_IMAGE_FORMAT_Y16:
			bit_depth = 16;
			color_type = PNG_COLOR_TYPE_GRAY;
			break;

		case FRAMEGRAB_IMAGE_FORMAT_RGB:
		default:
			bit_depth = 8;
			color_type = PNG_COLOR_TYPE_RGB;
			break;
	}

	png_set_IHDR(png_ptr, info_ptr, image->width, image->height, bit_depth, color_type,
			     PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

	/* Write the header info, the image, and end */
	png_write_info(png_ptr, info_ptr);
	png_write_image(png_ptr, row_ptrs);
	png_write_end(png_ptr, info_ptr);

	/* Clean up */
	free(row_ptrs);
	png_destroy_write_struct(&png_ptr, &info_ptr);

	if (file) {
		/* Close the file */
		fclose(file);
	} else {
		if (pbuffer && psize) {
			*pbuffer = (FG_UINT8*)buffer_data.buf;
			*psize   = (int)buffer_data.size;
		}
	}

	return 1;

png_write_error:

	if (row_ptrs) {
		free(row_ptrs);
	}

	if (png_ptr) {
		png_destroy_write_struct(&png_ptr, &info_ptr);
	}

	if (file) {
		fclose(file);
		remove(filename);
	}

	if (buffer_data.buf) {
		free(buffer_data.buf);
	}

	return 0;
}

/**
 * Writes an image to the specified file in the specified format.
 */
int php_framegrab_image_write(php_framegrab_image_data *image, char *filename, int format) {
	switch (format) {
		case FRAMEGRAB_IMAGE_OUTPUT_PNG:
			/* Some images we have to convert others we can output as is. */
			switch (image->format) {
				case FRAMEGRAB_IMAGE_FORMAT_RGB:
				case FRAMEGRAB_IMAGE_FORMAT_Y8:
				case FRAMEGRAB_IMAGE_FORMAT_Y16:
					return framegrab_image_write_to_png(image, filename, NULL, NULL);

				case FRAMEGRAB_IMAGE_FORMAT_YUYV:
				case FRAMEGRAB_IMAGE_FORMAT_UYVY:
				{
					php_framegrab_image_data newimage;

					if (framegrab_image_data_reformat(image, FRAMEGRAB_IMAGE_FORMAT_RGB, &newimage)) {
						int retval = framegrab_image_write_to_png(&newimage, filename, NULL, NULL);
						php_framegrab_image_data_destroy(&newimage);
						return retval;
					}
				}

				default:
					return 0;
			}

		/* TODO: Other formats */

		default:
			return 0;
	}
}

/**
 * Returns the image in the specified format in a buffer
 */
int php_framegrab_image_to_string(php_framegrab_image_data *image, FG_UINT8 **pbuffer, int *psize, int format) {
	switch (format) {
		case FRAMEGRAB_IMAGE_OUTPUT_PNG:
			/* Some images we have to convert others we can output as is. */
			switch (image->format) {
				case FRAMEGRAB_IMAGE_FORMAT_RGB:
				case FRAMEGRAB_IMAGE_FORMAT_Y8:
				case FRAMEGRAB_IMAGE_FORMAT_Y16:
					return framegrab_image_write_to_png(image, NULL, pbuffer, psize);

				case FRAMEGRAB_IMAGE_FORMAT_YUYV:
				case FRAMEGRAB_IMAGE_FORMAT_UYVY:
				{
					php_framegrab_image_data newimage;

					if (framegrab_image_data_reformat(image, FRAMEGRAB_IMAGE_FORMAT_RGB, &newimage)) {
						int retval = framegrab_image_write_to_png(&newimage, NULL, pbuffer, psize);
						php_framegrab_image_data_destroy(&newimage);
						return retval;
					}
				}

				default:
					return 0;
			}

		/* TODO: Other formats */

		default:
			return 0;
	}
}


/**
 * Returns a textual description of the image format.
 */
char *php_framegrab_image_format_to_string(int format) {
	switch (format) {
		case FRAMEGRAB_IMAGE_FORMAT_RGB:
			return "RGB";
		case FRAMEGRAB_IMAGE_FORMAT_Y8:
			return "Y8 (8 bit gray scale)";
		case FRAMEGRAB_IMAGE_FORMAT_Y16:
			return "Y16 (16 bit gray scale)";
		case FRAMEGRAB_IMAGE_FORMAT_YUYV:
			return "YUYV (4:2:2)";
		case FRAMEGRAB_IMAGE_FORMAT_UYVY:
			return "UYVY (4:2:2 reversed)";
		default:
			return "*UNKNOWN*";
	}
}

/**
 * Returns a character string describing the device.
 */
char *php_framegrab_image_data_to_string(php_framegrab_image_data *data) {
	char *buffer = (char*)calloc(FRAMEGRAB_IMAGE_STRING_BUFFER_SIZE, sizeof(char));

	snprintf(buffer, FRAMEGRAB_IMAGE_STRING_BUFFER_SIZE,
			 "FrameGrabImage:format=\"%s\",width=%d,height=%d",
			 php_framegrab_image_format_to_string(data->format), data->width, data->height);

	return buffer;
}

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * indent-tabs-mode: t
 * End:
 */
