/*
   +----------------------------------------------------------------------+
   | PHP Version 5                                                        |
   +----------------------------------------------------------------------+
   | Copyright (c) 1997-2008 The PHP Group                                |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.php.net/license/3_01.txt                                  |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
   | Author: Melanie Rhianna Lewis <cyberspice@php.net>                   |
   +----------------------------------------------------------------------+
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "../php_framegrab_device.h"
#include "../php_framegrab_image.h"

#if defined(__linux__)

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>

#include <linux/videodev2.h>

#define VIDEO_DEVICE "/dev/video%d"
#define FRAME_GRAB_TIMEOUT_SECS 120

/**
 * Does an ioctl, retrying if interrupted by a signal.
 */
static int fg_ioctl(int fd, int request, void * arg) {
	int ret;

	do {
		ret = ioctl(fd, request, arg);
	} while ((-1 == ret) && (EINTR == errno));

	return ret;
}

static void framegrab_init_device_with_caps(php_framegrab_device_data *data, int fd, struct v4l2_capability *caps) {
	struct v4l2_cropcap cropcaps;

	/* Initialise cropcaps. */
	memset(&cropcaps, 0, sizeof(cropcaps));
	cropcaps.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	/* Set default rectangle for crop caps just in case cropcaps
	 * isn't implemented. This is against the API but that's not
	 * stopped people before! */
	cropcaps.defrect.left   = 0;
	cropcaps.defrect.top    = 0;
	cropcaps.defrect.width  = 640;
	cropcaps.defrect.height = 480;

	/* Ignore the return value */
	(void)fg_ioctl(fd, VIDIOC_CROPCAP, &cropcaps);

	data->driver_name = (char*)calloc(strlen(caps->driver) + 1, sizeof(char));
	if (data->driver_name) {
		strcpy(data->driver_name, caps->driver);
	}

	data->card = (char*)calloc(strlen(caps->card) + 1, sizeof(char));
	if (data->card) {
		strcpy(data->card, caps->card);
	}

	data->def_left   = cropcaps.defrect.left;
	data->def_top    = cropcaps.defrect.top;
	data->def_width  = cropcaps.defrect.width;
	data->def_height = cropcaps.defrect.height;
}

/**
 * Returns the device data for the specified video device.
 */
int php_framegrab_get_device_data(php_framegrab_device_data *data, char *device) {
	struct v4l2_capability caps;
	int fd;

	/* Try and open the device and get basic info. */
	fd = open(device, O_RDWR|O_NDELAY);
	if (fd != -1) {
		if (!fg_ioctl(fd, VIDIOC_QUERYCAP, &caps)) {
			if ((caps.capabilities & V4L2_CAP_VIDEO_CAPTURE) == V4L2_CAP_VIDEO_CAPTURE) {

				/* Copy the name */
				data->device = (char*)calloc(strlen(device) + 1, sizeof(char));
				if (data->device) {
					strcpy(data->device, device);
				}

				framegrab_init_device_with_caps(data, fd, &caps);
				close(fd);
				return 1;

			} /* else not a capture device. */

		} /* else you have to at least support VIDIOC_QUERYCAP otherwise
		     you're a very broken driver. */
		close(fd);
	} /* else no such device */

	return 0;
}

/**
 * Returns the device data for the default video device.  The default video
 * device is the first one it finds which is capable of video capture.
 */
int php_framegrab_get_default_device_data(php_framegrab_device_data *data) {
	char device[32];
	struct v4l2_capability caps;
	int i, fd;

	/* Iterate through until we find a video device which both exists and
	 * supports video capture. */
	for (i = 0; ;i++) {
		snprintf(device, sizeof(device), VIDEO_DEVICE, i);

		fd = open(device, O_RDWR|O_NDELAY);
		if (fd != -1) {
			if (!fg_ioctl(fd, VIDIOC_QUERYCAP, &caps)) {
				if ((caps.capabilities & V4L2_CAP_VIDEO_CAPTURE) == V4L2_CAP_VIDEO_CAPTURE) {

					/* Copy the name */
					data->device = (char*)calloc(strlen(device) + 1, sizeof(char));
					if (data->device) {
						strcpy(data->device, device);
					}

					framegrab_init_device_with_caps(data, fd, &caps);
					close(fd);
					return 1;

				} /* else not a capture device, try again. */

			} /* else you have to at least support VIDIOC_QUERYCAP otherwise
			     you're a very broken driver. */
			close(fd);
			break;
		} else {
			break;
		}
	}

	return 0;
}

/**
 * Translates V4L2 format description in to a framegrab format description.
 */
static int framegrab_translate_format(int format) {
	switch (format) {
		case V4L2_PIX_FMT_GREY:
			return FRAMEGRAB_IMAGE_FORMAT_Y8;
		case V4L2_PIX_FMT_Y16:
			return FRAMEGRAB_IMAGE_FORMAT_Y16;
		case V4L2_PIX_FMT_YUYV:
			return FRAMEGRAB_IMAGE_FORMAT_YUYV;
		case V4L2_PIX_FMT_UYVY:
			return FRAMEGRAB_IMAGE_FORMAT_UYVY;
		default:
			return FRAMEGRAB_IMAGE_FORMAT_NONE;
	}
}

/**
 * Uses read() to grab a frame from the driver of the specified size
 * returning a pointer to a buffer containing the image data.
 */
static int framegrab_grab_image_with_read(int fd, int sizeimage, void **pbuf) {
	struct timeval tvs;
	fd_set fds;
	void *buffer;
	int ret;

	buffer = malloc(sizeimage);
	if (!buffer) {
		return 0;
	}

	/* Wait up to FRAME_GRAB_TIMEOUT_SECS seconds to grab a frame */
	tvs.tv_sec  = FRAME_GRAB_TIMEOUT_SECS;
	tvs.tv_usec = 0;

	/* The fd we're interest in */
    FD_ZERO (&fds);
    FD_SET (fd, &fds);

	ret = select(fd + 1, &fds, NULL, NULL, &tvs);
	if (1 == ret) { /* We have data */

		do {
			ret = read(fd, buffer, sizeimage);
		} while ((-1 == ret) && (EINTR == errno));

		*pbuf = buffer;
		return 1;
	} else if (-1 == ret) { /* We have an error */

		free(buffer);
		return 0;
	} else { /* We timed out */

		free(buffer);
		return 0;
	}
}

/**
 * Uses mmap() to grab a frame from the driver of the specified size
 * returning a pointer to a buffer containing the image data.
 */
static int framegrab_grab_image_with_mmap(int fd, int sizeimage, void **pbuf) {
	int retval = 0;
	struct timeval tv;
	fd_set fds;
	void *buffer;
	int length;
	struct v4l2_requestbuffers reqs;
	struct v4l2_buffer bufs;

	/* Request a buffer from the driver in which to grab an image */
	memset(&reqs, 0, sizeof(reqs));
    reqs.count  = 1;
    reqs.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    reqs.memory = V4L2_MEMORY_MMAP;

	if (-1 == fg_ioctl(fd, VIDIOC_REQBUFS, &reqs)) {
		return retval;
	}

	/* Must have at least one buffer */
	if (reqs.count < 1) {
		return retval;
	}

	/* Get the buffer's information */
	memset(&bufs, 0, sizeof(bufs));
	bufs.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	bufs.memory      = V4L2_MEMORY_MMAP;
	bufs.index       = 0;

	if (-1 == fg_ioctl(fd, VIDIOC_QUERYBUF, &bufs)) {
		return retval;
	}

	/* Map the buffer */
	length = bufs.length;
	buffer = mmap(NULL, length, PROT_READ|PROT_WRITE, MAP_SHARED, fd, bufs.m.offset);
	if (MAP_FAILED == buffer) {
		return retval;
	}

	/* Queue buffer, start stream, wait for a frame, dequeue and copy it. */
	if (!fg_ioctl(fd, VIDIOC_QBUF, &bufs)) {
		enum v4l2_buf_type type = bufs.type;

		if (!fg_ioctl(fd, VIDIOC_STREAMON, &type)) {
			struct timeval tvs;
			fd_set fds;
			int ret;

			/* Wait up to FRAME_GRAB_TIMEOUT_SECS seconds to grab a frame */
			tvs.tv_sec  = FRAME_GRAB_TIMEOUT_SECS;
			tvs.tv_usec = 0;

			/* The fd we're interest in */
			FD_ZERO (&fds);
			FD_SET (fd, &fds);

			ret = select(fd + 1, &fds, NULL, NULL, &tvs);
			if (1 == ret) {

				if (-1 == fg_ioctl(fd, VIDIOC_DQBUF, &bufs)) {

				} else {
					*pbuf = malloc(sizeimage);
					if (*pbuf) {
						memcpy(*pbuf, buffer, sizeimage);
						retval = 1;
					}
				}
			} else if (-1 == ret) {
				/* Error */
			} else {
				/* Timed out */
			}
			(void) fg_ioctl(fd, VIDIOC_STREAMOFF, &type);
		}
	}

	/* Unmap buffer */
	(void)munmap(buffer, length);

	return retval;
}

/**
 * Grabs an image from the specified device.
 */
int php_framegrab_image_grab(char *device, php_framegrab_image_data *image) {
	int retval = 0;
	struct v4l2_capability caps;
	struct v4l2_cropcap cropcaps;
	struct v4l2_crop crops;
	struct v4l2_format formats;
	void *imagedata;
	int fd;

	fd = open(device, O_RDWR|O_NDELAY);
	if (-1 == fd) {
		return retval;
	}

	if (!fg_ioctl(fd, VIDIOC_QUERYCAP, &caps)) {
		/* printf("Device  : %s\n", device);
		printf("Driver  : %s\n", caps.driver);
		printf("Card    : %s\n", caps.card);
		printf("Bus info: %s\n", caps.bus_info);
		printf("Caps    : 0x%08x\n", caps.capabilities); */

		if ((caps.capabilities & V4L2_CAP_VIDEO_CAPTURE) == V4L2_CAP_VIDEO_CAPTURE) {
			memset(&cropcaps, 0, sizeof(cropcaps));
			cropcaps.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

			/* Set default rectangle for crop caps just in case cropcaps
			 * isn't implemented. This is against the API but that's not
			 * stopped people before! */
			cropcaps.defrect.left   = 0;
			cropcaps.defrect.top    = 0;
			cropcaps.defrect.width  = 640;
			cropcaps.defrect.height = 480;

			/* Ignore the return value */
			(void)fg_ioctl(fd, VIDIOC_CROPCAP, &cropcaps);

			/* Crop to the default cropping rectangle */
			memset(&crops, 0, sizeof(crops));
			crops.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
			crops.c    = cropcaps.defrect;

			/* printf("left=%d, top=%d, width=%d, height=%d\n",
					crops.c.left, crops.c.top, crops.c.width, crops.c.height); */

			/* Ignore the return value */
			(void)fg_ioctl(fd, VIDIOC_S_CROP, &crops);

			memset(&formats, 0, sizeof(formats));
			formats.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
			formats.fmt.pix.width       = crops.c.width;
			formats.fmt.pix.height      = crops.c.height;
			formats.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
			formats.fmt.pix.field       = V4L2_FIELD_INTERLACED;

			if (!fg_ioctl(fd, VIDIOC_S_FMT, &formats)) {
				/* printf("width=%d, height=%d, bytesperline=%u, sizeimage=%u\n",
						formats.fmt.pix.width, formats.fmt.pix.height,
						formats.fmt.pix.bytesperline, formats.fmt.pix.sizeimage); */

				if ((caps.capabilities & V4L2_CAP_STREAMING) == V4L2_CAP_STREAMING) {
					retval = framegrab_grab_image_with_mmap(fd, formats.fmt.pix.sizeimage, &imagedata);
				} else if ((caps.capabilities & V4L2_CAP_READWRITE) == V4L2_CAP_READWRITE) {
					retval = framegrab_grab_image_with_read(fd, formats.fmt.pix.sizeimage, &imagedata);
				}

				if (retval) {
					image->bytes_per_line = formats.fmt.pix.bytesperline;
					image->format         = framegrab_translate_format(formats.fmt.pix.pixelformat);
					image->width          = formats.fmt.pix.width;
					image->height         = formats.fmt.pix.height;
					image->size           = formats.fmt.pix.sizeimage;
					image->image          = imagedata;
					close(fd);
					return 1;
				}
			}
		} else {
			/* Video capture not supported */
		}
	}
	close(fd);

	return 0;
}

#else

#error Linux only supported at the moment!  Sorry!

#endif

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * indent-tabs-mode: t
 * End:
 */
