/*
   +----------------------------------------------------------------------+
   | PHP Version 5                                                        |
   +----------------------------------------------------------------------+
   | Copyright (c) 1997-2010 The PHP Group                                |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.php.net/license/3_01.txt                                  |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
   | Author: Melanie Rhianna Lewis <cyberspice@php.net>                   |
   +----------------------------------------------------------------------+
 */

#ifndef PHP_FRAMEGRAB_IMAGE_H_
#define PHP_FRAMEGRAB_IMAGE_H_

/**
 * The type for an 8 bit unsigned value.
 */
#define FG_UINT8 unsigned char

/**
 * The image contains no data.
 */
#define FRAMEGRAB_IMAGE_FORMAT_NONE    0

/**
 * The image contains data as an RGB image with three bytes per pixel in
 * RGB order.  The R byte is at the lower memory address compared to the B
 * byte.
 */
#define FRAMEGRAB_IMAGE_FORMAT_RGB     1

/**
 * The image contains data as a grey scale image with one byte per pixel.
 */
#define FRAMEGRAB_IMAGE_FORMAT_Y8      2

/**
 * The image contains data as a grey scale image with two bytes per pixel
 * in little endian format (i.e. the low byte is at the lower memory address
 * compared to the high byte).
 */
#define FRAMEGRAB_IMAGE_FORMAT_Y16     3

/**
 * The image contains data as a colour image with four bytes per pixel pair
 * in the following format Y Cb Y Cr.  I.e. the horizontal colour resolution
 * is half the horizontal luminance resolution.  The vertical resolutions
 * are the same.  This is also know as YUV 4:2:2 (or YUV2 on Windows).
 */
#define FRAMEGRAB_IMAGE_FORMAT_YUYV    4

/**
 * This is as above but with the byte order or pairs of bytes reverse.
 * I.e. Cb Y Cr Y.
 */
#define FRAMEGRAB_IMAGE_FORMAT_UYVY    5

/**
 * Output the image as PNG
 */
#define FRAMEGRAB_IMAGE_OUTPUT_PNG     1

/**
 * Structure that contains the information that describes a block of image
 * data.
 */
typedef struct _php_framegrab_image_data {
	int format;
	int width;
	int height;
	int bytes_per_line;
	int size;
	void *image;

} php_framegrab_image_data;

/**
 * Initialises a php_framegrab_image_data structure.
 */
void php_framegrab_image_data_init(php_framegrab_image_data *data);

/**
 * Destroys a php_framegrab_image_data structure freeing any memory
 * it is using.
 */
void php_framegrab_image_data_destroy(php_framegrab_image_data *data);

/**
 * Clones a php_framegrab_image_data structure.
 */
int php_framegrab_image_data_clone(php_framegrab_image_data *data,
		                           php_framegrab_image_data *copydata);

/**
 * Grabs an image from the specified device.
 */
int php_framegrab_image_grab(char *device, php_framegrab_image_data *image);

/**
 * Writes an image to the specified file in the specified format.
 */
int php_framegrab_image_write(php_framegrab_image_data *image, char *filename, int format);

/**
 * Returns the image in the specified format in a buffer
 */
int php_framegrab_image_to_string(php_framegrab_image_data *image, FG_UINT8 **pbuffer, int *psize, int format);

/**
 * Returns a textual description of the image format.
 */
char *php_framegrab_image_format_to_string(int format);

/**
 * Returns a character string describing the device.
 */
char *php_framegrab_image_data_to_string(php_framegrab_image_data *data);


#endif /* PHP_FRAMEGRAB_IMAGE_H_ */

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * indent-tabs-mode: t
 * End:
 */
